####################################################################
# Example Neural Net code for Survey Practice special issue
#
# Adam Eck, 10/27/17
####################################################################
rm(list=ls())
library(mxnet)

######################################################
############ LOAD AND PREPROCESS THE DATA ############
######################################################

# load the data set
setwd("/local/scratch/SurveyPractice/models")
load("/home/aeck/Dropbox/MachineLearningPapers/DataSets/New.RData")
PaperData <- as.data.frame(SPData2[, 1])
names(PaperData)[1] <- "id"

# preprocess the data

# one hot code all of the categorical attributes
# region + sex + hispanic2 + wborace + educ3 + wrkcata + telstat + incgrp4 + age + ratcat2
PaperData$region_NE <- as.numeric(as.numeric(SPData2$region) == 1)
PaperData$region_MW <- as.numeric(as.numeric(SPData2$region) == 2)
PaperData$region_S <- as.numeric(as.numeric(SPData2$region) == 3)

PaperData$sex <- as.numeric(SPData2$sex) - 1

PaperData$hispanic <- as.numeric(SPData2$hispanic2) - 1

PaperData$race_W <- as.numeric(as.numeric(SPData2$wborace) == 1)
PaperData$race_B <- as.numeric(as.numeric(SPData2$wborace) == 2)

PaperData$educ_HS <- as.numeric(as.numeric(SPData2$educ3) == 1)
PaperData$educ_SC <- as.numeric(as.numeric(SPData2$educ3) == 2)

PaperData$wrkcata_PCE <- as.numeric(as.numeric(SPData2$wrkcata) == 1)
PaperData$wrkcata_FGE <- as.numeric(as.numeric(SPData2$wrkcata) == 2)
PaperData$wrkcata_SGE <- as.numeric(as.numeric(SPData2$wrkcata) == 3)
PaperData$wrkcata_LGE <- as.numeric(as.numeric(SPData2$wrkcata) == 4)
PaperData$wrkcata_SE <- as.numeric(as.numeric(SPData2$wrkcata) == 5)
PaperData$wrkcata_NWE <- as.numeric(as.numeric(SPData2$wrkcata) == 6)

PaperData$telstat_LL <- as.numeric(as.numeric(SPData2$telstat) == 1 | as.numeric(SPData2$telstat) == 2)
PaperData$telstat_Cell <- as.numeric(as.numeric(SPData2$telstat) == 3 | as.numeric(SPData2$telstat) == 2)

PaperData$incgrp_0 <- as.numeric(as.numeric(SPData2$incgrp4) == 1)
PaperData$incgrp_35 <- as.numeric(as.numeric(SPData2$incgrp4) == 2)
PaperData$incgrp_50 <- as.numeric(as.numeric(SPData2$incgrp4) == 3)

# normalize the numeric attributes
PaperData$age <- (SPData2$age - mean(SPData2$age)) / sd(SPData2$age)

PaperData$ratcat2 <- (SPData2$ratcat2 - mean(SPData2$ratcat2)) / sd(SPData2$ratcat2)

# add the labels
PaperData$newrespond20 <- as.numeric(SPData2$newrespond20) - 1

# add the fold information
PaperData$fold <- SPData2$fold

#############################################################
############ CREATE TRAINING AND TEST DATA SETS #############
#############################################################

# set the RNG seed
set.seed(12345)

# create the training and test sets
testFold = max(PaperData$fold)
labelIndex = ncol(PaperData) - 1

test.x = data.matrix(PaperData[PaperData$fold == testFold, 2:(labelIndex - 1)])
test.y = PaperData[PaperData$fold == testFold, c(labelIndex)]

train.x = data.matrix(PaperData[PaperData$fold < testFold, 2:(labelIndex - 1)])
train.y = PaperData[PaperData$fold < testFold, c(labelIndex)]

#########################################
############ EXPERIMENTAL SETUP #########
#########################################

# set the seed for training
mx.set.seed(12345)

# assign the parameters
numLayers = 1
hiddenNeurons = c(2, 5, 10, 20, 50, 100)
epochs = 200
learningRate = 0.1

#########################################
############ CREATE THE MODELS ##########
#########################################

for (numNeurons in hiddenNeurons) {
  # create the input layer
  data <- mx.symbol.Variable("data")
  
  # create the hidden layer
  net <- mx.symbol.FullyConnected(data, num.hidden=numNeurons)
  act <- mx.symbol.Activation(data = net, act_type="sigmoid")
  
  # create the output layer
  outNet <- mx.symbol.FullyConnected(act, num.hidden=2)
  mlp <- mx.symbol.SoftmaxOutput(outNet)    
  
  # train and save the model
  modelPrefix = paste0("newrespond20_", numNeurons, "n")
  model <- mx.model.FeedForward.create(symbol = mlp,
                 X = train.x,
                 y = train.y,
                 array.layout = "rowmajor",
                 num.round = epochs,
                 learning.rate = learningRate,
                 eval.metric = mx.metric.accuracy,
                 epoch.end.callback = mx.callback.save.checkpoint(modelPrefix, period = 1),
                 verbose = FALSE)
}

#########################################
############ EVALUTE THE MODEL ##########
#########################################

# create the lists for the results
index <- 1
neuronsList <- list()
epochsList <- list()
accList <- list()
sensList <- list()
specList <- list()

# load each trained model for evaluation
for (numNeurons in hiddenNeurons) {
  for (epoch in 1:epochs) {
    # load the model
    modelPrefix = paste0("newrespond20_", numNeurons, "n")
    model = mx.model.load(modelPrefix, iteration = epoch)
    
    # make the predictions on the test set
    predicted.raw <- predict(model, test.x, array.layout = "rowmajor")
    predicted.values <- max.col(t(predicted.raw)) - 1
    
    # calcuate the confusion matrix
    results <- data.frame(test.y, predicted.values)
    names(results) <- c("actual", "predicted")
    conMatrix <- table(results$actual, results$predicted)
    
    # calculate the evaluation metrics
    tn <- ifelse(is.na(conMatrix[1]), 0, conMatrix[1])
    fn <- ifelse(is.na(conMatrix[2]), 0, conMatrix[2])
    fp <- ifelse(is.na(conMatrix[3]), 0, conMatrix[3])
    tp <- ifelse(is.na(conMatrix[4]), 0, conMatrix[4])
    
    acc <- (tp + tn) / (tp + fp + fn + tn)
    sens <- tp / (tp + fn)
    spec <- tn / (tn + fp)
    
    # save the results
    neuronsList[[index]] <- numNeurons
    epochsList[[index]] <- epoch
    accList[[index]] <- acc
    sensList[[index]] <- sens
    specList[[index]] <- spec
    index <- index + 1
    
    # print(paste0("N: ", numNeurons, " E: ", epoch, " Acc: ", acc, " Sens: ", sens, " Spec: ", spec))
    # print(paste0("TP: ", tp, " FP: ", fp, " FN: ", fn, " TN: ", tn))
  }
}

#########################################
############ SAVE THE RESULTS ###########
#########################################

nnResults <- data.frame(unlist(neuronsList), unlist(epochsList), unlist(accList), unlist(sensList), unlist(specList))
names(nnResults) <- c("Model", "Epoch", "Accuracy", "Sensitivity", "Specificity")

filename <- paste0("/home/aeck/Dropbox/MachineLearningPapers/Neural/R/results_mxnet_newrespond20.csv")
write.csv(nnResults, filename, na="", row.names=FALSE)

#########################################
############ RUN LOGITS #################
#########################################

# reset the RNG seed
set.seed(12345)

# split the data again
train = SPData2[SPData2$fold < testFold,]
test = SPData2[SPData2$fold == testFold,]

# train the logistic regression
# incgrp4 needs to be converted to factor or it will be treated as ordinal.
regression <- glm(newrespond20 ~ region + sex + hispanic2 + wborace + educ3 + wrkcata 
                  + telstat + as.factor(incgrp4) + age + ratcat2,
                  data = train,
                  family="binomial")

# compute the results over the test set
confusionMatrix.lr <- array(0, dim=c(2,2))
lr.results <- predict(regression, newdata = test, type="response")
results <- data.frame(actual = as.numeric(test$newrespond20) - 1,
                      predictionRaw = lr.results,
                      prediction = as.numeric(lr.results >= 0.5),
                      id = test$id)

for (j in 1:nrow(results)) {
  if (!is.na(results$actual[j])) {
    confusionMatrix.lr[results$actual[j] + 1, results$prediction[j] + 1] <- confusionMatrix.lr[results$actual[j] + 1, results$prediction[j] + 1] + 1
  }
}

# process the results
n.lr <- confusionMatrix.lr[1, 1] + confusionMatrix.lr[1, 2] + confusionMatrix.lr[2, 1] + confusionMatrix.lr[2, 2]
accuracy.lr <- (confusionMatrix.lr[1, 1] + confusionMatrix.lr[2, 2]) / n.lr
sens.lr <- confusionMatrix.lr[2, 2] / (confusionMatrix.lr[2, 1] + confusionMatrix.lr[2, 2])
spec.lr <- confusionMatrix.lr[1, 1] / (confusionMatrix.lr[1, 1] + confusionMatrix.lr[1, 2])

#########################################
############ COMBINE THE RESULTS ########
#########################################

results <- nnResults
results$Model <- paste0("NN", results$Model)

for (epoch in 1:epochs) {
  results <- rbind(results, c("LR", epoch, accuracy.lr, sens.lr, spec.lr))
}

results$Model <- factor(results$Model, levels = c("LR", "NN2", "NN5", "NN10", "NN20", "NN50", "NN100"))
results$Epoch <- as.integer(results$Epoch)
results$Accuracy <- as.numeric(results$Accuracy)
results$Sensitivity <- as.numeric(results$Sensitivity)
results$Specificity <- as.numeric(results$Specificity)

results$Balanced <- results$Sensitivity * results$Specificity

#########################################
############ PLOT THE RESULTS ###########
#########################################

library(ggplot2)
ggplot(data=results[seq(10, nrow(results), 10),], aes(x=Epoch, y=Accuracy, group=Model, colour=Model, shape=Model)) + 
  scale_shape_manual(values = c(2000, 1:6)) +
  geom_line() + 
  geom_point() +
  scale_x_continuous(breaks = scales::pretty_breaks(n=10)) +
  scale_y_continuous(breaks = scales::pretty_breaks(n=10)) +
  labs(x = "Training Iterations") +
  labs(title = "Predictive Accuracy of Models") +
  labs(caption = "Results for Logistic Regression (LR) drawn as a straight line for comparison") +
  theme(plot.title = element_text(hjust = 0.5), text=element_text(size=16))

ggplot(data=results[seq(10, nrow(results), 10),], aes(x=Epoch, y=Sensitivity, group=Model, colour=Model, shape=Model)) + 
  scale_shape_manual(values = c(2000, 1:6)) +
  geom_line() + 
  geom_point() +
  scale_x_continuous(breaks = scales::pretty_breaks(n=10)) +
  scale_y_continuous(breaks = scales::pretty_breaks(n=10)) +
  labs(x = "Training Iterations") +
  labs(title = "Predictive Sensitivity of Models") +
  labs(caption = "Results for Logistic Regression (LR) drawn as a straight line for comparison") +
  theme(plot.title = element_text(hjust = 0.5), text=element_text(size=16))

ggplot(data=results[seq(10, nrow(results), 10),], aes(x=Epoch, y=Specificity, group=Model, colour=Model, shape=Model)) + 
  scale_shape_manual(values = c(2000, 1:6)) +
  geom_line() + 
  geom_point() +
  scale_x_continuous(breaks = scales::pretty_breaks(n=10)) +
  scale_y_continuous(breaks = scales::pretty_breaks(n=10)) +
  labs(x = "Training Iterations") +
  labs(title = "Predictive Specificity of Models") +
  labs(caption = "Results for Logistic Regression (LR) drawn as a straight line for comparison") +
  theme(plot.title = element_text(hjust = 0.5), text=element_text(size=16))

ggplot(data=results[seq(10, nrow(results), 10),], aes(x=Epoch, y=Balanced, group=Model, colour=Model, shape=Model)) + 
  scale_shape_manual(values = c(2000, 1:6)) +
  geom_line() + 
  geom_point() +
  scale_x_continuous(breaks = scales::pretty_breaks(n=10)) +
  scale_y_continuous(breaks = scales::pretty_breaks(n=10)) +
  labs(x = "Training Iterations") +
  labs(y = "Balanced Accuracy") +
  labs(title = "Balanced Predictive Accuracy of Models") +
  labs(caption = "Results for Logistic Regression (LR) drawn as a straight line for comparison") +
  theme(plot.title = element_text(hjust = 0.5), text=element_text(size=16))
